#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from functools import partial
from pathlib import Path

from kiwi.data.corpus import Corpus
from kiwi.data.fieldsets import extend_vocabs_fieldset
from kiwi.data.fieldsets.fieldset import Fieldset
from kiwi.data.qe_dataset import QEDataset
from kiwi.data.utils import (
    build_vocabulary,
    filter_len,
    load_vocabularies_to_datasets,
)


def build_dataset(fieldset, prefix='', filter_pred=None, **kwargs):
    fields, files = fieldset.fields_and_files(prefix, **kwargs)
    examples = Corpus.from_files(fields=fields, files=files)
    dataset = QEDataset(
        examples=examples, fields=fields, filter_pred=filter_pred
    )
    return dataset


def build_training_datasets(
    fieldset,
    split=0.0,
    valid_source=None,
    valid_target=None,
    load_vocab=None,
    **kwargs,
):
    """Build a training and validation QE datasets.

    Required Args:
        fieldset (Fieldset): specific set of fields to be used (depends on
                             the model to be used).
        train_source: Train Source
        train_target: Train Target (MT)

    Optional Args (depends on the model):
        train_pe: Train Post-edited
        train_target_tags: Train Target Tags
        train_source_tags: Train Source Tags
        train_sentence_scores: Train HTER scores

        valid_source: Valid Source
        valid_target: Valid Target (MT)
        valid_pe: Valid Post-edited
        valid_target_tags: Valid Target Tags
        valid_source_tags: Valid Source Tags
        valid_sentence_scores: Valid HTER scores

        split (float): If no validation sets are provided, randomly sample
                       1 - split of training examples as validation set.

        target_vocab_size: Maximum Size of target vocabulary
        source_vocab_size: Maximum Size of source vocabulary
        target_max_length: Maximum length for target field
        target_min_length: Minimum length for target field
        source_max_length: Maximum length for source field
        source_min_length: Minimum length for source field
        target_vocab_min_freq: Minimum word frequency target field
        source_vocab_min_freq: Minimum word frequency source field
        load_vocab: Path to existing vocab file

    Returns:
        A training and a validation Dataset.
    """
    # TODO: improve handling these length options (defaults are set multiple
    # times).
    filter_pred = partial(
        filter_len,
        source_min_length=kwargs.get('source_min_length', 1),
        source_max_length=kwargs.get('source_max_length', float('inf')),
        target_min_length=kwargs.get('target_min_length', 1),
        target_max_length=kwargs.get('target_max_length', float('inf')),
    )
    train_dataset = build_dataset(
        fieldset, prefix=Fieldset.TRAIN, filter_pred=filter_pred, **kwargs
    )
    if valid_source and valid_target:
        valid_dataset = build_dataset(
            fieldset,
            prefix=Fieldset.VALID,
            filter_pred=filter_pred,
            valid_source=valid_source,
            valid_target=valid_target,
            **kwargs,
        )
    elif split:
        if not 0.0 < split < 1.0:
            raise Exception(
                'Invalid data split value: {}; it must be in the '
                '(0, 1) interval.'.format(split)
            )
        train_dataset, valid_dataset = train_dataset.split(split)
    else:
        raise Exception('Validation data not provided.')

    if load_vocab:
        vocab_path = Path(load_vocab)
        load_vocabularies_to_datasets(vocab_path, train_dataset, valid_dataset)

    # Even if vocab is loaded, we need to build the vocabulary
    # in case fields are missing
    datasets_for_vocab = [train_dataset]
    if kwargs.get('extend_source_vocab') or kwargs.get('extend_target_vocab'):
        vocabs_fieldset = extend_vocabs_fieldset.build_fieldset(fieldset)
        extend_vocabs_ds = build_dataset(vocabs_fieldset, **kwargs)
        datasets_for_vocab.append(extend_vocabs_ds)

    fields_vocab_options = fieldset.fields_vocab_options(**kwargs)
    build_vocabulary(fields_vocab_options, *datasets_for_vocab)

    return train_dataset, valid_dataset


def build_test_dataset(fieldset, load_vocab=None, **kwargs):
    """Build a test QE dataset.

    Args:
      fieldset (Fieldset): specific set of fields to be used (depends on
                           the model to be used.)
      load_vocab: A path to a saved vocabulary.

    Returns:
        A Dataset object.
    """

    test_dataset = build_dataset(fieldset, prefix=Fieldset.TEST, **kwargs)

    fields_vocab_options = fieldset.fields_vocab_options(**kwargs)
    if load_vocab:
        vocab_path = Path(load_vocab)
        load_vocabularies_to_datasets(vocab_path, test_dataset)
    else:
        build_vocabulary(fields_vocab_options, test_dataset)

    return test_dataset



"""This implements a linear model."""
#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from .sparse_vector import SparseVector


class LinearModel(object):
    """ An abstract linear model."""

    def __init__(self):
        self.use_average = True
        self.weights = SparseVector()
        self.averaged_weights = SparseVector()

    def clear(self):
        """Clear all weights."""
        self.weights.clear()
        self.averaged_weights.clear()

    def finalize(self, t):
        """Finalize by setting the weights as the running average.
        This is a no-op if use_average=False."""
        if self.use_average:
            self.averaged_weights.scale(1.0 / float(t))
            self.weights.add(self.averaged_weights)

    def compute_score(self, features):
        """Compute a score by taking the inner product with a feature
        vector."""
        score = features.dot_product(self.weights)
        return score

    def compute_score_binary_features(self, binary_features):
        """Compute a score by taking the inner product with a binary
        feature vector."""
        score = 0.0
        for f in binary_features:
            if f in self.weights:
                score += self.weights[f]
        return score

    def make_gradient_step(self, features, eta, t, gradient):
        """Make a gradient step with stepsize eta."""
        self.weights.add(features, -eta * gradient)
        if self.use_average:
            self.averaged_weights.add(features, eta * float(t) * gradient)

    def save(self, model_file, average=False, feature_indices=None):
        """Save the model to a file."""
        f = open(model_file, 'w')
        if feature_indices is not None:
            w = SparseVector()
            for index in self.weights:
                w[feature_indices.get_label_name(index)] = self.weights[index]
            w.save(f)
        else:
            self.weights.save(f)
        f.close()
        if average:
            f = open(model_file + '_average', 'w')
            if feature_indices is not None:
                w = SparseVector()
                for index in self.averaged_weights:
                    w[
                        feature_indices.get_label_name(index)
                    ] = self.averaged_weights[index]
                w.save(f)
            else:
                self.averaged_weights.save(f)
            f.close()

    def load(self, model_file, average=False, feature_indices=None):
        """Load the model from a file."""
        f = open(model_file, 'r')
        if feature_indices is not None:
            w = SparseVector()
            w.load(f)
            for key in w:
                index = feature_indices.add(key)
                self.weights[index] = w[key]
        else:
            self.weights.load(f)
        f.close()
        if average:
            f = open(model_file + '_average', 'r')
            if feature_indices is not None:
                w = SparseVector()
                w.load(f)
                for key in w:
                    index = feature_indices.get_label_id(key)
                    self.averaged_weights[index] = w[key]
            else:
                self.averaged_weights.load(f)
            f.close()

    def write_fnames(self, fnames_file, fnames):
        """Write file mapping from integers to feature descriptions."""
        f = open(fnames_file, 'w')
        for fid, fname in enumerate(fnames):
            f.write(str(1 + fid) + ' ' + fname + '\n')
        f.close()

    def read_fnames(self, fnames_file):
        """Read file mapping from integers to feature descriptions."""
        assert False, 'This is not being called'
        fids = {}
        f = open(fnames_file)
        maxfid = -1
        for line in f:
            line = line.rstrip('\n')
            fields = line.split(' ')
            fid = int(fields[0])
            fname = fields[1]
            fids[fname] = fid
            if fid > maxfid:
                maxfid = fid
        fnames = [''] * maxfid
        for fname, fid in fids.iteritems():
            fnames[fid - 1] = fname
        f.close()
        return fnames, fids



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import functools
import logging
from collections import OrderedDict

from kiwi.loggers import tracking_logger

logger = logging.getLogger(__name__)


@functools.total_ordering
class StatsSummary(OrderedDict):
    def __init__(self, prefix=None, main_metric=None, ordering=max, **kwargs):
        self.prefix = prefix
        self._main_metric_name = main_metric
        self.ordering = ordering
        super().__init__(**kwargs)

    @property
    def main_metric(self):
        if self._main_metric_name:
            return self._main_metric_name
        elif self:
            return list(self.keys())[0]
        return None

    def main_metric_value(self):
        return self.__getitem__(self.main_metric)

    def _make_key(self, key):
        if self.prefix:
            key = '{}_{}'.format(self.prefix, key)
        return key

    def __str__(self):
        return ', '.join(['{}: {:0.4f}'.format(k, v) for k, v in self.items()])

    def log(self):
        """Log statistics to output and also to tracking logger.

        :param stats_summary: StatsSummary object
        """
        print('\r', end='\r')
        logger.info(self)

        for k, v in self.items():
            tracking_logger.log_metric(k, v)

    def __setitem__(self, key, value):
        key = self._make_key(key)
        super().__setitem__(key, value)

    def __getitem__(self, key):
        key = self._make_key(key)
        return super().__getitem__(key)

    def __contains__(self, key):
        key = self._make_key(key)
        return super().__contains__(key)

    def get(self, key, default=None):
        key = self._make_key(key)
        return super().get(key, default)

    def __eq__(self, other):
        return isinstance(other, StatsSummary) and self.get(
            self.main_metric
        ) == other.get(self.main_metric)

    def __le__(self, other):
        if self.ordering == max:
            return isinstance(other, StatsSummary) and self.get(
                self.main_metric
            ) <= other.get(self.main_metric)
        else:
            return isinstance(other, StatsSummary) and self.get(
                self.main_metric
            ) >= other.get(self.main_metric)

    def __gt__(self, other):
        if self.ordering == max:
            return isinstance(other, StatsSummary) and self.get(
                self.main_metric
            ) > other.get(self.main_metric)
        else:
            return isinstance(other, StatsSummary) and self.get(
                self.main_metric
            ) < other.get(self.main_metric)

    def better_than(self, other):
        if self.ordering == max:
            return isinstance(other, StatsSummary) and self.get(
                self.main_metric
            ) > other.get(self.main_metric)
        else:
            return isinstance(other, StatsSummary) and self.get(
                self.main_metric
            ) < other.get(self.main_metric)


class Stats:
    def __init__(
        self,
        metrics,
        main_metric=None,
        main_metric_ordering=max,
        log_interval=0,
    ):
        self.metrics = metrics
        main_metric = main_metric or self.metrics[0]
        self.main_metric_name = main_metric.get_name()
        self.main_metric_ordering = main_metric_ordering

        self.log_interval = log_interval

        self.reset()

    def update(self, **kwargs):
        self.steps += 1
        for metric in self.metrics:
            metric.update(**kwargs)

    def summarize(self, prefix=None):
        summary = StatsSummary(
            prefix=prefix,
            main_metric=self.main_metric_name,
            ordering=self.main_metric_ordering,
        )
        if self.steps:
            for metric in self.metrics:
                summary.update(metric.summarize())
        return summary

    def reset(self):
        self.steps = 0
        for metric in self.metrics:
            metric.reset()

    def wrap_up(self, prefix=None):
        summary = self.summarize(prefix)
        self.reset()
        return summary

    def log(self, step=None):
        if (
            step is None
            or self.log_interval > 0
            and not step % self.log_interval
        ):
            stats_summary = self.wrap_up()
            stats_summary.log()



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse
from distutils.util import strtobool
from pathlib import Path

from kiwi import constants as const


class PathType(object):
    """Factory for creating pathlib.Path objects

    Instances of PathType should passed as type= arguments to the
    ArgumentParser add_argument() method.

    Strongly based on argparse.FileType.

    Keyword Arguments:
        - exists -- Whether the file must exists or not.

    """

    def __init__(self, exists=False):
        self._must_exist = exists

    def __call__(self, string):
        if not string:
            return string

        # The special argument "-" means sys.std{in,out} in argparse.FileType
        if string == '-':
            msg = (
                "argument type PathType does not support '-' for referring "
                "to sys.std{in,out}"
            )
            raise ValueError(msg)

        # all other arguments are used as file names
        path = Path(string)
        if self._must_exist and not path.exists():
            message = 'path must exist: {}'.format(string)
            raise argparse.ArgumentTypeError(message)
        return str(path)

    def __repr__(self):
        arg_str = repr(self._must_exist)
        return '{}({})'.format(type(self).__name__, arg_str)


def io_opts(parser):
    # Logging
    group = parser.add_argument_group('I/O')
    group.add_argument(
        '--save-config',
        required=False,
        type=PathType(exists=False),
        is_write_out_config_file_arg=False,
        # Setting it to true makes it save and exit
        help='Save parsed configuration and arguments to the specified file',
    )
    group.add_argument(
        '-d', '--debug', action='store_true', help='Output additional messages.'
    )
    group.add_argument(
        '-q',
        '--quiet',
        action='store_true',
        help='Only output warning and error messages.',
    )


def logging_opts(parser):
    # Logging options
    group = parser.add_argument_group('Logging')
    group.add_argument(
        '--log-interval', type=int, default=100, help='Log every k batches.'
    )
    group.add_argument(
        '--mlflow-tracking-uri',
        type=str,
        default='mlruns/',
        help='If using MLflow, logs model parameters, training metrics, and '
        'artifacts (files) to this MLflow server. Uses the localhost by '
        'default.',
    )
    group.add_argument(
        '--experiment-name',
        required=False,
        help='If using MLflow, it will log this run under this experiment '
        'name, which appears as a separate section'
        'in the UI. It will also be used in some messages and files.',
    )
    group.add_argument(
        '--run-name',
        required=False,
        help='If using MLflow, it will log this run under this run '
        'name, which appears as a separate item in the experiment.',
    )
    group.add_argument(
        '--run-uuid',
        required=False,
        help='If specified, MLflow/Default Logger will log metrics and params '
        'under this ID. If it exists, the run status will '
        'change to running. This ID is also used for creating '
        'this run\'s output directory. '
        '(Run ID must be a 32-character hex string)',
    )
    group.add_argument(
        '--output-dir',
        type=str,
        help='Output several files for this run under this directory. '
        'If not specified, a directory under "runs" is created '
        'or reused based on the Run UUID. '
        'Files might also be sent to MLflow depending on the '
        '--mlflow-always-log-artifacts option.',
    )
    group.add_argument(
        '--mlflow-always-log-artifacts',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='If using MLFlow, always log (send) artifacts (files) to MLflow '
        'artifacts URI. By default (false), artifacts are only logged if'
        'MLflow is a remote server (as specified by --mlflow-tracking-uri '
        'option). All generated files are always saved in --output-dir, so it '
        'might be considered redundant to copy them to a local MLflow '
        'server. If this is not the case, set this option to true.',
    )


def general_opts(parser):
    # Data processing options
    group = parser.add_argument_group('random')
    group.add_argument('--seed', type=int, default=42, help='Random seed')

    # Cuda
    group = parser.add_argument_group('gpu')
    group.add_argument(
        '--gpu-id',
        default=None,
        type=int,
        help='Use CUDA on the listed devices',
    )


def save_load_opts(parser):
    group = parser.add_argument_group('save-load')
    group.add_argument(
        '--load-model',
        type=PathType(exists=True),
        help='Directory containing a {} file to be loaded'.format(
            const.MODEL_FILE
        ),
    )
    group.add_argument(
        '--save-data',
        type=str,
        help='Output dir for saving the preprocessed data files.',
    )
    group.add_argument(
        '--load-data',
        type=PathType(exists=True),
        help='Input dir for loading the preprocessed data files.',
    )

    group.add_argument(
        '--load-vocab',
        type=PathType(exists=True),
        help='Directory containing a {} file to be loaded'.format(
            const.VOCAB_FILE
        ),
    )



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from functools import partial

from kiwi.data.vectors import AvailableVectors


class Fieldset:
    ALL = 'all'
    TRAIN = 'train'
    VALID = 'valid'
    TEST = 'test'

    def __init__(self):
        """

        """
        self._fields = {}
        self._options = {}
        self._required = {}
        self._vocab_options = {}
        self._vocab_vectors = {}
        self._file_reader = {}

    def add(
        self,
        name,
        field,
        file_option_suffix,
        required=ALL,
        vocab_options=None,
        vocab_vectors=None,
        file_reader=None,
    ):
        """

        Args:
            name:
            field:
            file_option_suffix:
            required (str or list or None):
            file_reader (callable): by default, uses Corpus.from_files().

        Returns:

        """
        self._fields[name] = field
        self._options[name] = file_option_suffix
        if not isinstance(required, list):
            required = [required]
        self._required[name] = required
        self._file_reader[name] = file_reader

        if vocab_options is None:
            vocab_options = {}
        self._vocab_options[name] = vocab_options
        self._vocab_vectors[name] = vocab_vectors

    @property
    def fields(self):
        return self._fields

    def is_required(self, name, set_name):
        required = self._required[name]
        if set_name in required or self.ALL in required:
            return True
        else:
            return False

    def fields_and_files(self, set_name, **files_options):
        fields = {}
        files = {}
        for name, file_option_suffix in self._options.items():
            file_option = '{}{}'.format(set_name, file_option_suffix)
            file_name = files_options.get(file_option)
            if not file_name and self.is_required(name, set_name):
                raise FileNotFoundError(
                    'File {} is required (use the {} '
                    'option).'.format(file_name, file_option.replace('_', '-'))
                )
            elif file_name:
                files[name] = {
                    'name': file_name,
                    'reader': self._file_reader.get(name),
                }
                fields[name] = self._fields[name]
        return fields, files

    # def files_formats(self):
    #     return {
    #         set_name: self._file_format.get(set_name)
    #         for set_name in self._fields
    #     }
    #
    def vocab_kwargs(self, name, **kwargs):
        if name not in self._vocab_options:
            raise KeyError(
                'Field named "{}" does not exist in this fieldset'.format(name)
            )
        vkwargs = {}
        for argument, option_name in self._vocab_options[name].items():
            option_value = kwargs.get(option_name)
            if option_value is not None:
                vkwargs[argument] = option_value
        return vkwargs

    def vocab_vectors_loader(
        self,
        name,
        embeddings_format='polyglot',
        embeddings_binary=False,
        **kwargs
    ):
        if name not in self._vocab_vectors:
            raise KeyError(
                'Field named "{}" does not exist in this fieldset'.format(name)
            )

        def no_vectors_fn():
            return None

        vectors_fn = no_vectors_fn

        option_name = self._vocab_vectors[name]
        if option_name:
            option_value = kwargs.get(option_name)
            if option_value:
                emb_model = AvailableVectors[embeddings_format]
                # logger.info('Loading {} embeddings from {}'.format(
                #     name, option_value))
                vectors_fn = partial(
                    emb_model, option_value, binary=embeddings_binary
                )
        return vectors_fn

    def vocab_vectors(self, name, **kwargs):
        vectors_fn = self.vocab_vectors_loader(name, **kwargs)
        return vectors_fn()

    def fields_vocab_options(self, **kwargs):
        vocab_options = {}
        for name, field in self.fields.items():
            vocab_options[name] = dict(
                vectors_fn=self.vocab_vectors_loader(name, **kwargs)
            )
            vocab_options[name].update(self.vocab_kwargs(name, **kwargs))
        return vocab_options



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn

import kiwi
from kiwi import constants as const
from kiwi.data import utils
from kiwi.models.utils import load_torch_file

logger = logging.getLogger(__name__)


class ModelConfig:
    __metaclass__ = ABCMeta

    def __init__(self, vocabs):
        """Model Configuration Base Class.

        Args:
        vocabs: Dictionary Mapping Field Names to Vocabularies.
                Must contain 'source' and 'target' keys
        """
        self.source_vocab_size = len(vocabs[const.SOURCE])
        self.target_vocab_size = len(vocabs[const.TARGET])

    @classmethod
    def from_dict(cls, config_dict, vocabs):
        """Create config from a saved state_dict.
           Args:
             config_dict: A dictionary that is the return value of
                          a call to the `state_dict()` method of `cls`
             vocab: See `ModelConfig.__init__`
        """
        config = cls(vocabs)
        config.update(config_dict)
        return config

    def update(self, other_config):
        """Updates the config object with the values of `other_config`
           Args:
             other_config: The `dict` or `ModelConfig` object to update with.
        """
        config_dict = dict()
        if isinstance(self, other_config.__class__):
            config_dict = other_config.__dict__
        elif isinstance(other_config, dict):
            config_dict = other_config
        self.__dict__.update(config_dict)

    def state_dict(self):
        """Return the __dict__ for serialization.
        """
        self.__dict__['__version__'] = kiwi.__version__
        return self.__dict__


class Model(nn.Module):
    __metaclass__ = ABCMeta

    subclasses = {}

    def __init__(self, vocabs, ConfigCls=ModelConfig, config=None, **kwargs):
        """Quality Estimation Base Class.

        Args:
            vocabs: Dictionary Mapping Field Names to Vocabularies.
            ConfigCls: ModelConfig Subclass
            config: A State Dict of a ModelConfig subclass.
                    If set, passing other kwargs will raise an Exception.
        """

        super().__init__()

        self.vocabs = vocabs

        if config is None:
            config = ConfigCls(vocabs=vocabs, **kwargs)
        else:
            config = ConfigCls.from_dict(config_dict=config, vocabs=vocabs)
            assert not kwargs
        self.config = config

    @classmethod
    def register_subclass(cls, subclass):
        cls.subclasses[subclass.__name__] = subclass
        return subclass

    @abstractmethod
    def loss(self, model_out, target):
        pass

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass

    def num_parameters(self):
        return sum(p.numel() for p in self.parameters())

    def predict(self, batch, class_name=const.BAD, unmask=True):
        model_out = self(batch)
        predictions = {}
        class_index = torch.tensor([const.LABELS.index(class_name)])

        for key in model_out:
            if key in [const.TARGET_TAGS, const.SOURCE_TAGS, const.GAP_TAGS]:
                # Models are assumed to return logits, not probabilities
                logits = model_out[key]
                probs = torch.softmax(logits, dim=-1)
                class_probs = probs.index_select(
                    -1, class_index.to(device=probs.device)
                )
                class_probs = class_probs.squeeze(-1).tolist()
                if unmask:
                    if key == const.SOURCE_TAGS:
                        input_key = const.SOURCE
                    else:
                        input_key = const.TARGET
                    mask = self.get_mask(batch, input_key)
                    if key == const.GAP_TAGS:
                        # Append one extra token
                        mask = torch.cat(
                            [mask.new_ones((mask.shape[0], 1)), mask], dim=1
                        )

                    lengths = mask.int().sum(dim=-1)
                    for i, x in enumerate(class_probs):
                        class_probs[i] = x[: lengths[i]]
                predictions[key] = class_probs
            elif key == const.SENTENCE_SCORES:
                predictions[key] = model_out[key].tolist()
            elif key == const.BINARY:
                logits = model_out[key]
                probs = torch.softmax(logits, dim=-1)
                class_probs = probs.index_select(
                    -1, class_index.to(device=probs.device)
                )
                predictions[key] = class_probs.tolist()

        return predictions

    def predict_raw(self, examples):
        batch = self.preprocess(examples)
        return self.predict(batch, class_name=const.BAD_ID, unmask=True)

    def preprocess(self, examples):
        """Preprocess Raw Data.

        Args:
            examples (list of dict): List of examples. Each Example is a dict
                                     with field strings as keys, and
                                     unnumericalized, tokenized data as values.
        Return:
            A batch object.
        """
        raise NotImplementedError

    def get_mask(self, batch, output):
        """Compute Mask of Tokens for side.

        Args:
            batch: Namespace of tensors
            side: String identifier.
        """
        side = output
        # if output in [const.TARGET_TAGS, const.GAP_TAGS]:
        #     side = const.TARGET
        # elif output == const.SOURCE_TAGS:
        #     side = const.SOURCE

        input_tensor = getattr(batch, side)
        if isinstance(input_tensor, tuple) and len(input_tensor) == 2:
            input_tensor, lengths = input_tensor

        # output_tensor = getattr(batch, output)
        # if isinstance(output_tensor, tuple) and len(output_tensor) == 2:
        #     output_tensor, lengths = output_tensor

        mask = torch.ones_like(input_tensor, dtype=torch.uint8)

        possible_padding = [const.PAD, const.START, const.STOP]
        unk_id = self.vocabs[side].stoi.get(const.UNK)
        for pad in possible_padding:
            pad_id = self.vocabs[side].stoi.get(pad)
            if pad_id is not None and pad_id != unk_id:
                mask &= torch.as_tensor(
                    input_tensor != pad_id,
                    device=mask.device,
                    dtype=torch.uint8,
                )

        return mask

    @staticmethod
    def create_from_file(path):

        try:
            model_dict = load_torch_file(path)
        except FileNotFoundError:
            # If no model is found
            raise FileNotFoundError(
                'No valid model data found in {}'.format(path)
            )

        for model_name in Model.subclasses:
            if model_name in model_dict:
                model = Model.subclasses[model_name].from_dict(model_dict)
                return model

    @classmethod
    def from_file(cls, path):
        model_dict = torch.load(
            str(path), map_location=lambda storage, loc: storage
        )
        if cls.__name__ not in model_dict:
            raise KeyError(
                '{} model data not found in {}'.format(cls.__name__, path)
            )

        return cls.from_dict(model_dict)

    @classmethod
    def from_dict(cls, model_dict):
        vocabs = utils.deserialize_vocabs(model_dict[const.VOCAB])
        class_dict = model_dict[cls.__name__]
        model = cls(vocabs=vocabs, config=class_dict[const.CONFIG])
        model.load_state_dict(class_dict[const.STATE_DICT])
        return model

    def save(self, path):
        vocabs = utils.serialize_vocabs(self.vocabs)
        model_dict = {
            '__version__': kiwi.__version__,
            const.VOCAB: vocabs,
            self.__class__.__name__: {
                const.CONFIG: self.config.state_dict(),
                const.STATE_DICT: self.state_dict(),
            },
        }
        torch.save(model_dict, str(path))



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from distutils.util import strtobool

from kiwi import constants as const
from kiwi.cli.better_argparse import ModelParser
from kiwi.cli.opts import PathType
from kiwi.lib.utils import parse_integer_with_positive_infinity
from kiwi.models.predictor_estimator import Estimator

title = 'Estimator (Predictor-Estimator)'


def _add_training_data_file_opts(parser):
    # Data options
    group = parser.add_argument_group('data')

    group.add_argument(
        '--train-source',
        type=PathType(exists=True),
        required=True,
        help='Path to training source file',
    )
    group.add_argument(
        '--train-target',
        type=PathType(exists=True),
        # required=True,
        help='Path to training target file',
    )
    group.add_argument(
        '--train-source-tags',
        type=PathType(exists=True),
        help='Path to validation label file for source (WMT18 format)',
    )
    group.add_argument(
        '--train-target-tags',
        type=PathType(exists=True),
        help='Path to validation label file for target',
    )

    group.add_argument(
        '--train-pe',
        type=PathType(exists=True),
        help='Path to file containing post-edited target.',
    )
    group.add_argument(
        '--train-sentence-scores',
        type=PathType(exists=True),
        help='Path to file containing sentence level scores.',
    )

    valid_group = parser.add_argument_group('validation data')

    valid_group.add_argument(
        '--split',
        type=float,
        help='Split Train dataset in case that no validation set is given.',
    )

    valid_group.add_argument(
        '--valid-source',
        type=PathType(exists=True),
        # required=True,
        help='Path to validation source file',
    )
    valid_group.add_argument(
        '--valid-target',
        type=PathType(exists=True),
        # required=True,
        help='Path to validation target file',
    )
    valid_group.add_argument(
        '--valid-alignments',
        type=str,
        # required=True,
        help='Path to valid alignments between source and target.',
    )
    valid_group.add_argument(
        '--valid-source-tags',
        type=PathType(exists=True),
        help='Path to validation label file for source (WMT18 format)',
    )
    valid_group.add_argument(
        '--valid-target-tags',
        type=PathType(exists=True),
        help='Path to validation label file for target',
    )

    valid_group.add_argument(
        '--valid-pe',
        type=PathType(exists=True),
        help='Path to file containing postedited target.',
    )
    valid_group.add_argument(
        '--valid-sentence-scores',
        type=PathType(exists=True),
        help='Path to file containing sentence level scores.',
    )


def _add_predicting_data_file_opts(parser):
    # Data options
    group = parser.add_argument_group('data')

    group.add_argument(
        '--test-source',
        type=PathType(exists=True),
        required=True,
        help='Path to validation source file',
    )
    group.add_argument(
        '--test-target',
        type=PathType(exists=True),
        required=True,
        help='Path to validation target file',
    )
    return group


def _add_data_flags(parser):
    group = parser.add_argument_group('data processing options')
    group.add_argument(
        '--predict-side',
        type=str,
        default=const.TARGET_TAGS,
        choices=[const.TARGET_TAGS, const.SOURCE_TAGS, const.GAP_TAGS],
        help='Tagset to predict. Leave unchanged for WMT17 format.',
    )
    group.add_argument(
        '--wmt18-format',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Read target tags in WMT18 format.',
    )

    group.add_argument(
        '--source-max-length',
        type=parse_integer_with_positive_infinity,
        default=float("inf"),
        help='Maximum source sequence length',
    )
    group.add_argument(
        '--source-min-length',
        type=int,
        default=0,
        help='Truncate source sequence length.',
    )
    group.add_argument(
        '--target-max-length',
        type=parse_integer_with_positive_infinity,
        default=float("inf"),
        help='Maximum target sequence length to keep.',
    )
    group.add_argument(
        '--target-min-length',
        type=int,
        default=0,
        help='Truncate target sequence length.',
    )

    return group


def _add_vocabulary_opts(parser):
    group = parser.add_argument_group(
        'vocabulary options',
        description='Options for loading vocabulary from a previous run. '
        'This is used for e.g. training a source predictor via predict-'
        'inverse: True ; If set, other vocab options are ignored',
    )
    group.add_argument(
        '--source-vocab-size',
        type=int,
        default=None,
        help='Size of the source vocabulary.',
    )
    group.add_argument(
        '--target-vocab-size',
        type=int,
        default=None,
        help='Size of the target vocabulary.',
    )
    group.add_argument(
        '--source-vocab-min-frequency',
        type=int,
        default=1,
        help='Min word frequency for source vocabulary.',
    )
    group.add_argument(
        '--target-vocab-min-frequency',
        type=int,
        default=1,
        help='Min word frequency for target vocabulary.',
    )


def _add_data_options(data_parser):
    group = data_parser.add_argument_group(
        'PredEst data',
        description='Predictor Estimator specific data ' 'options. (POSTECH)',
    )

    group.add(
        '--extend-source-vocab',
        type=PathType(exists=True),
        help='Optionally load more data which is used only for vocabulary '
        'creation. Path to additional Data'
        '(Predictor)',
    )
    group.add(
        '--extend-target-vocab',
        type=PathType(exists=True),
        help='Optionally load more data which is used only for vocabulary '
        'creation. Path to additional Data'
        '(Predictor)',
    )


def add_pretraining_options(parser):
    _add_training_data_file_opts(parser)
    _add_data_flags(parser)
    _add_vocabulary_opts(parser)
    _add_data_options(parser)

    group = parser.add_argument_group(
        'predictor training', description='Predictor Estimator (POSTECH)'
    )
    # Only for training
    group.add_argument(
        '--warmup',
        type=int,
        default=0,
        help='Pretrain Predictor for this number of steps.',
    )
    group.add_argument(
        '--rnn-layers-pred', type=int, default=2, help='Layers in Pred RNN'
    )
    group.add_argument(
        '--dropout-pred', type=float, default=0.0, help='Dropout in predictor'
    )
    group.add_argument(
        '--hidden-pred',
        type=int,
        default=100,
        help='Size of hidden layers in LSTM',
    )
    group.add_argument(
        '--out-embeddings-size',
        type=int,
        default=200,
        help='Word Embedding in Output layer',
    )
    group.add_argument(
        '--embedding-sizes',
        type=int,
        default=0,
        help='If set, takes precedence over other embedding params',
    )
    group.add_argument(
        '--share-embeddings',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Tie input and output embeddings for target.',
    )

    group.add_argument(
        '--predict-inverse',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict target -> source instead of source -> target.',
    )

    group = parser.add_argument_group(
        'model-embeddings',
        description='Embedding layers size in case pre-trained embeddings '
        'are not used.',
    )
    group.add_argument(
        '--source-embeddings-size',
        type=int,
        default=50,
        help='Word embedding size for source.',
    )
    group.add_argument(
        '--target-embeddings-size',
        type=int,
        default=50,
        help='Word embedding size for target.',
    )


def add_training_options(training_parser):
    add_pretraining_options(training_parser)

    group = training_parser.add_argument_group(
        'predictor-estimator training',
        description='Predictor Estimator (POSTECH). These settings are used '
        ' to train the Predictor. They will be ignored if training a '
        ' Predictor-Estimator and the `load-model` flag is set.',
    )
    group.add_argument(
        '--start-stop',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Append start and stop symbols to estimator feature sequence.',
    )
    group.add_argument(
        '--predict-gaps',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict Gap Tags. Requires `train-gap-tags`, `valid-'
        'gap-tags` to be set.',
    )
    group.add_argument(
        '--predict-target',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=True,
        help='Predict Target Tags. Requires `train-target-tags`, `valid-'
        'target-tags` to be set.',
    )
    group.add_argument(
        '--predict-source',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict Source Tags. Requires `train-source-tags`, `valid-'
        'source-tags` to be set.',
    )
    group.add_argument(
        '--load-pred-source',
        type=PathType(exists=True),
        help='If set, model architecture and vocabulary parameters are '
        'ignored. Load pretrained predictor tgt->src.',
    )
    group.add_argument(
        '--load-pred-target',
        type=PathType(exists=True),
        help='If set, model architecture and vocabulary parameters are '
        'ignored. Load pretrained predictor src->tgt.',
    )

    group.add_argument(
        '--rnn-layers-est', type=int, default=2, help='Layers in Estimator RNN'
    )
    group.add_argument(
        '--dropout-est', type=float, default=0.0, help='Dropout in estimator'
    )
    group.add_argument(
        '--hidden-est',
        type=int,
        default=100,
        help='Size of hidden layers in LSTM',
    )
    group.add_argument(
        '--mlp-est',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help="""Pass the Estimator input through a linear layer
        reducing dimensionality before RNN.""",
    )
    group.add_argument(
        '--sentence-level',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help="""Predict Sentence Level Scores.
        Requires setting `train-sentence-scores, valid-sentence-scores`""",
    )
    group.add_argument(
        '--sentence-ll',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help="""Use probabilistic Loss for sentence scores instead of
        squared error. If set, the model will output mean and variance of
        a truncated Gaussian distribution over the interval [0, 1], and use
        the NLL of ground truth `hter` as the loss.
        This seems to improve performance, and gives you uncertainty estimates
        for sentence level predictions as a byproduct.
        If `sentence-level == False`, this is without effect.
        """,
    )
    group.add_argument(
        '--sentence-ll-predict-mean',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help="""If `sentence-ll == True`, by default the prediction for `hter`
        will be the mean of the Guassian /before truncation/. After truncation,
        this will be the mode of the distribution, but not the mean as
        truncated Gaussian is skewed to one side. set this to `True` to use
        the True mean after truncation for prediction.
        """,
    )
    group.add_argument(
        '--use-probs',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict scores as product/sum of word level probs',
    )

    group.add_argument(
        '--binary-level',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help="""Predict binary sentence labels indicating `hter == 0.0`
        Requires setting `train-sentence-scores`, `valid-sentence-scores`""",
    )
    group.add_argument(
        '--token-level',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help="""Continue training the predictor on the postedited text.
        If set, will do an additional forward pass through the predictor
        Using the SRC, PE pair and add the `Predictor` loss for the tokens
        in the postedited text PE. Recommended if you have access to PE.
        Requires setting `train-pe`, `valid-pe`""",
    )
    group.add_argument(
        '--target-bad-weight',
        type=float,
        default=3.0,
        help='Relative weight for target bad labels.',
    )
    group.add_argument(
        '--gaps-bad-weight',
        type=float,
        default=3.0,
        help='Relative weight for gaps bad labels.',
    )
    group.add_argument(
        '--source-bad-weight',
        type=float,
        default=3.0,
        help='Relative weight for source bad labels.',
    )


def add_predicting_options(predicting_parser):
    _add_predicting_data_file_opts(predicting_parser)
    group = predicting_parser.add_argument_group(
        'predictor-estimator Prediction',
        description='Predictor Estimator (POSTECH)',
    )
    group.add_argument(
        '--wmt18-format',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Read target tags in WMT18 format.',
    )

    group.add_argument(
        '--sentence-level',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict Sentence Level Scores',
    )
    group.add_argument(
        '--binary-level',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict binary sentence labels',
    )
    group.add_argument(
        '--valid-batch-size', type=int, default=32, help='Batch Size'
    )
    group.add_argument(
        '--predict-inverse',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict target -> source instead of source -> target.',
    )


def parser_for_pipeline(pipeline):
    if pipeline == 'train':
        return ModelParser(
            'estimator',
            'train',
            title=Estimator.title,
            options_fn=add_training_options,
            api_module=Estimator,
        )
    if pipeline == 'predict':
        return ModelParser(
            'estimator',
            'predict',
            title=Estimator.title,
            options_fn=add_predicting_options,
            api_module=Estimator,
        )

    return None



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import torch
from torchtext import data

from kiwi import constants as const
from kiwi.data import utils
from kiwi.data.fields.sequence_labels_field import SequenceLabelsField
from kiwi.data.fieldsets.fieldset import Fieldset
from kiwi.data.tokenizers import tokenizer


def build_text_field():
    return data.Field(
        tokenize=tokenizer,
        init_token=const.START,
        batch_first=True,
        eos_token=const.STOP,
        pad_token=const.PAD,
        unk_token=const.UNK,
    )


def build_label_field(postprocessing=None):
    return SequenceLabelsField(
        classes=const.LABELS,
        tokenize=tokenizer,
        pad_token=const.PAD,
        batch_first=True,
        postprocessing=postprocessing,
    )


def build_fieldset(wmt18_format=False):
    target_field = build_text_field()
    source_field = build_text_field()

    source_vocab_options = dict(
        min_freq='source_vocab_min_frequency', max_size='source_vocab_size'
    )
    target_vocab_options = dict(
        min_freq='target_vocab_min_frequency', max_size='target_vocab_size'
    )

    fieldset = Fieldset()
    fieldset.add(
        name=const.SOURCE,
        field=source_field,
        file_option_suffix='_source',
        required=Fieldset.TRAIN,
        vocab_options=source_vocab_options,
    )
    fieldset.add(
        name=const.TARGET,
        field=target_field,
        file_option_suffix='_target',
        required=Fieldset.TRAIN,
        vocab_options=target_vocab_options,
    )
    fieldset.add(
        name=const.PE,
        field=target_field,
        file_option_suffix='_pe',
        required=None,
        vocab_options=target_vocab_options,
    )

    post_pipe_target = data.Pipeline(utils.project)
    if wmt18_format:
        post_pipe_gaps = data.Pipeline(utils.wmt18_to_gaps)
        post_pipe_target = data.Pipeline(utils.wmt18_to_target)
        fieldset.add(
            name=const.GAP_TAGS,
            field=build_label_field(post_pipe_gaps),
            file_option_suffix='_target_tags',
            required=None,
        )

    fieldset.add(
        name=const.TARGET_TAGS,
        field=build_label_field(post_pipe_target),
        file_option_suffix='_target_tags',
        required=None,
    )
    fieldset.add(
        name=const.SOURCE_TAGS,
        field=build_label_field(),
        file_option_suffix='_source_tags',
        required=None,
    )
    fieldset.add(
        name=const.SENTENCE_SCORES,
        field=data.Field(
            sequential=False, use_vocab=False, dtype=torch.float32
        ),
        file_option_suffix='_sentence_scores',
        required=None,
    )

    pipe = data.Pipeline(utils.hter_to_binary)
    fieldset.add(
        name=const.BINARY,
        field=data.Field(
            sequential=False,
            use_vocab=False,
            dtype=torch.long,
            preprocessing=pipe,
        ),
        file_option_suffix='_sentence_scores',
        required=None,
    )
    return fieldset



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from collections import defaultdict

import torch
from torchtext.data import Example

from kiwi import constants as const
from kiwi.data.iterators import build_bucket_iterator
from kiwi.data.qe_dataset import QEDataset

logger = logging.getLogger(__name__)


class Predicter:
    def __init__(self, model, fields=None):
        """Class to load a model for inference.

        Args:
          model (kiwi.models.Model): A trained QE model
          fields (dict[str: Field]): A dict mapping field names to strings.
            For online prediction.
        """

        self.model = model
        self.fields = fields
        # Will break in Multi GPU mode
        self._device = next(model.parameters()).device

    def predict(self, examples, batch_size=1):
        """Create Predictions for a list of examples.

           Args:
             examples: A dict  mapping field names to the
               list of raw examples (strings).
             batch_size: Batch Size to use. Default 1.

           Returns:
             A dict mapping prediction levels
             (word, sentence ..) to the model predictions
             for each example.

           Raises:
             Exception: If an example has an empty string
               as `source` or `target` field.

           Example:
             >>> import kiwi
             >>> predictor = kiwi.load_model('tests/toy-data/models/nuqe.torch')
             >>> src = ['a b c', 'd e f g']
             >>> tgt = ['q w e r', 't y']
             >>> align = ['0-0 1-1 1-2', '1-1 3-0']
             >>> examples = [kiwi.constants.SOURCE: src,
                             kiwi.constants.TARGET: tgt,
                             kiwi.constants.ALIGNMENTS: align]
             >>> predictor.predict(examples)
             {'tags': [[0.4760947525501251,
                0.47569847106933594,
                0.4948718547821045,
                0.5305878520011902],
               [0.5105430483818054, 0.5252899527549744]]}
        """
        if not examples:
            return defaultdict(list)
        if self.fields is None:
            raise Exception('Missing fields object.')

        if not examples.get(const.SOURCE):
            raise KeyError('Missing required field "{}"'.format(const.SOURCE))
        if not examples.get(const.TARGET):
            raise KeyError('Missing required field "{}"'.format(const.TARGET))

        if not all(
            [s.strip() for s in examples[const.SOURCE] + examples[const.TARGET]]
        ):
            raise Exception(
                'Empty String in {} or {} field found!'.format(
                    const.SOURCE, const.TARGET
                )
            )
        fields = [(name, self.fields[name]) for name in examples]

        field_examples = [
            Example.fromlist(values, fields)
            for values in zip(*examples.values())
        ]

        dataset = QEDataset(field_examples, fields=fields)

        return self.run(dataset, batch_size)

    def run(self, dataset, batch_size=1):
        iterator = build_bucket_iterator(
            dataset, self._device, batch_size, is_train=False
        )
        self.model.eval()
        predictions = defaultdict(list)
        with torch.no_grad():
            for batch in iterator:
                model_pred = self.model.predict(batch)
                for key, values in model_pred.items():
                    if isinstance(values, list):
                        predictions[key] += values
                    else:
                        predictions[key].append(values)
        return dict(predictions)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np
from more_itertools import collapse

# def calibrate_threshold(scores, labels, MetricClass=LazyF1):
#     """Finds optimal decision threshold according to metric.
#     Args:
#         scores (list[float]): List of model output scores
#         labels (list): List of corresponding target labels
#     Returns:
#     (metric, threshold): The value of the Metric and the Threshold to be used.
#     """
#     metric = MetricClass(scores, labels)
#     scores, labels = metric.sort(scores, labels)
#     init_threshold = scores[0]
#     thresholds = [(metric.compute(), init_threshold)]
#     for score, label in zip(scores, labels):
#         metric.update(score, label)
#         thresholds.append((metric.compute(), score))
#     return metric.choose(thresholds)


def mean_absolute_error(y, y_hat):
    return np.mean(np.absolute(y_hat - y))


def mean_squared_error(y, y_hat):
    return np.square(np.subtract(y, y_hat)).mean()


def delta_average(y_true, y_rank):
    """Calculate the DeltaAvg score

    This is a much faster version than the Perl one provided in the
    WMT QE task 1.

    References: could not find any.

    Author: Fabio Kepler (contributed to MARMOT)

    Args:
        y_true: array of reference score (not rank) of each segment.
        y_rank: array of rank of each segment.

    Returns: the absolute delta average score.

    """
    sorted_ranked_indexes = np.argsort(y_rank)
    y_length = len(sorted_ranked_indexes)

    delta_avg = 0
    max_quantiles = y_length // 2
    set_value = (
        np.sum(y_true[sorted_ranked_indexes[np.arange(y_length)]]) / y_length
    )
    quantile_values = {
        head: np.sum(y_true[sorted_ranked_indexes[np.arange(head)]]) / head
        for head in range(2, y_length)
    }
    # Cache values, since there are many that are repeatedly computed
    # between various quantiles.
    for quantiles in range(2, max_quantiles + 1):  # Current number of quantiles
        quantile_length = y_length // quantiles
        quantile_sum = 0
        for head in np.arange(
            quantile_length, quantiles * quantile_length, quantile_length
        ):
            quantile_sum += quantile_values[head]
        delta_avg += quantile_sum / (quantiles - 1) - set_value

    if max_quantiles > 1:
        delta_avg /= max_quantiles - 1
    else:
        delta_avg = 0
    return abs(delta_avg)


def precision(tp, fp, fn):
    if tp + fp > 0:
        return tp / (tp + fp)
    return 0


def recall(tp, fp, fn):
    if tp + fn > 0:
        return tp / (tp + fn)
    return 0


def fscore(tp, fp, fn):
    p = precision(tp, fp, fn)
    r = recall(tp, fp, fn)
    if p + r > 0:
        return 2 * (p * r) / (p + r)
    return 0


def confusion_matrix(hat_y, y, n_classes=None):
    hat_y = np.array(list(collapse(hat_y)))
    y = np.array(list(collapse(y)))

    if n_classes is None:
        classes = np.unique(np.union1d(hat_y, y))
        n_classes = len(classes)

    cnfm = np.zeros((n_classes, n_classes))
    for j in range(y.shape[0]):
        cnfm[y[j], hat_y[j]] += 1
    return cnfm


def scores_for_class(class_index, cnfm):
    tp = cnfm[class_index, class_index]
    fp = cnfm[:, class_index].sum() - tp
    fn = cnfm[class_index, :].sum() - tp
    tn = cnfm.sum() - tp - fp - fn

    p = precision(tp, fp, fn)
    r = recall(tp, fp, fn)
    f1 = fscore(tp, fp, fn)
    support = tp + tn
    return p, r, f1, support


def precision_recall_fscore_support(hat_y, y, labels=None):
    n_classes = len(labels) if labels else None
    cnfm = confusion_matrix(hat_y, y, n_classes)

    if n_classes is None:
        n_classes = cnfm.shape[0]

    scores = np.zeros((n_classes, 4))
    for class_id in range(n_classes):
        scores[class_id] = scores_for_class(class_id, cnfm)
    return scores.T.tolist()


def f1_product(hat_y, y):
    p, r, f1, s = precision_recall_fscore_support(hat_y, y)
    f1_mult = np.prod(f1)
    return f1_mult


def f1_scores(hat_y, y):
    """
    Return f1_bad, f1_ok and f1_product
    """
    p, r, f1, s = precision_recall_fscore_support(hat_y, y)
    f_mult = np.prod(f1)

    return (*f1, f_mult)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kiwi',
 'kiwi.cli',
 'kiwi.cli.models',
 'kiwi.cli.pipelines',
 'kiwi.data',
 'kiwi.data.fields',
 'kiwi.data.fieldsets',
 'kiwi.lib',
 'kiwi.metrics',
 'kiwi.models',
 'kiwi.models.linear',
 'kiwi.models.modules',
 'kiwi.predictors',
 'kiwi.trainers']

package_data = \
{'': ['*']}

install_requires = \
['configargparse>=0.14.0,<0.15.0',
 'more-itertools>=5.0,<6.0',
 'numpy>=1.16,<2.0',
 'pyyaml>=3.13,<4.0',
 'scipy>=1.2,<2.0',
 'torch>=0.4.1',
 'torchtext>=0.3.1,<0.4.0',
 'tqdm>=4.29,<5.0']

extras_require = \
{':python_version == "3.5"': ['pathlib2>=2.3,<3.0'],
 'embeddings': ['polyglot>=16.7,<17.0'],
 'mlflow': ['mlflow>=0.8,<0.9'],
 'plots': ['seaborn>=0.9.0,<0.10.0']}

entry_points = \
{'console_scripts': ['kiwi = kiwi.__main__:main']}

setup_kwargs = {
    'name': 'openkiwi',
    'version': '0.1.3',
    'description': 'Machine Translation Quality Estimation Toolkit',
    'long_description': "![OpenKiwi Logo](https://github.com/Unbabel/OpenKiwi/blob/master/docs/_static/img/openkiwi-logo-horizontal.svg)\n\n--------------------------------------------------------------------------------\n\n[![PyPI version](https://img.shields.io/pypi/v/openkiwi?color=%236ecfbd&label=pypi%20package&style=flat-square)](https://pypi.org/project/openkiwi/)\n[![python versions](https://img.shields.io/pypi/pyversions/openkiwi.svg?style=flat-square)](https://pypi.org/project/openkiwi/)\n[![CircleCI](https://img.shields.io/circleci/build/github/Unbabel/OpenKiwi/master?style=flat-square)](https://circleci.com/gh/Unbabel/OpenKiwi/tree/master)\n[![Code Climate coverage](https://img.shields.io/codeclimate/coverage/Unbabel/OpenKiwi?style=flat-square)](https://codeclimate.com/github/Unbabel/OpenKiwi/test_coverage)\n[![Code Style](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square)](https://github.com/psf/black)\n[![GitHub last commit](https://img.shields.io/github/last-commit/unbabel/openkiwi?style=flat-square)](https://github.com/Unbabel/OpenKiwi/commits/master)\n\n**Open-Source Machine Translation Quality Estimation in PyTorch**\n\nQuality estimation (QE) is one of the missing pieces of machine translation: its goal is to evaluate a translation system’s quality without access to reference translations. We present **OpenKiwi**, a Pytorch-based open-source framework that implements the best QE systems from WMT 2015-18 shared tasks, making it easy to experiment with these models under the same framework. Using OpenKiwi and a stacked combination of these models we have achieved state-of-the-art results on word-level QE on the WMT 2018 English-German dataset.\n\n## News\n\nFollowing our nomination in early July, we are happy to announce we won the [Best Demo Paper at ACL 2019](http://www.acl2019.org/EN/winners-of-acl-2019-best-paper-awards.xhtml)! Congratulations to the whole team and huge thanks for supporters and issue reporters.\n\nCheck out the [published paper](https://www.aclweb.org/anthology/P19-3020).\n\nWe are going to release the web interface we had put in place for the live demo presentation at ACL.\n\nWe have released the OpenKiwi [tutorial](https://github.com/Unbabel/KiwiCutter) we presented at MT Marathon 2019.\n\n## Features\n\n* Framework for training QE models and using pre-trained models for evaluating MT.\n* Supports both word and sentence-level Quality estimation.\n* Implementation of five QE systems in Pytorch: QUETCH [[1]], NuQE [[2], [3]], predictor-estimator [[4], [5]], APE-QE [[3]], and a stacked ensemble with a linear system [[2], [3]]. \n* Easy to use API. Import it as a package in other projects or run from the command line.\n* Provides scripts to run pre-trained QE models on data from the WMT 2018 campaign.\n* Easy to track and reproduce experiments via yaml configuration files.\n\n## Results\n\nResults for the [WMT18 Quality Estimation shared task](http://www.statmt.org/wmt18/quality-estimation-task.html#results), for [word level](https://competitions.codalab.org/competitions/19306#results) *and* [sentence level](https://competitions.codalab.org/competitions/19316#results) on the test set.\n\n|   Model   | En-De SMT |           |           |           |           | En-De NMT |           |           |           |           |\n|:---------:|:---------:|:---------:|:---------:|:---------:|:---------:|:---------:|:---------:|:---------:|:---------:|:---------:|\n|           |     MT    |    gaps   |   source  |     r     |     ⍴     |     MT    |    gaps   |   source  |     r     |     ⍴     |\n|  OpenKiwi | **62.70** | **52.14** | **48.88** |   71.08   |   72.70   | **44.77** | **22.89** | **36.53** |   46.72   |   58.51   |\n|  Wang2018 |   62.46   |   49.99   |     --    | **73.97** | **75.43** |   43.61   |     --    |     --    |   50.12   |   60.49   |\n|    UNQE   |     --    |     --    |     --    |   70.00   |   72.44   |     --    |     --    |     --    | **51.29** | **60.52** |\n| deepQUEST |   42.98   |   28.24   |   33.97   |   48.72   |   50.97   |   30.31   |   11.93   |   28.59   |   38.08   |   48.00   |\n\n\n## Quick Installation\n\nTo install OpenKiwi as a package, simply run\n```bash\npip install openkiwi\n```\n\nYou can now\n```python\nimport kiwi\n```\ninside your project or run in the command line\n```bash\nkiwi\n```\n\n**Optionally**, if you'd like to take advantage of our [MLflow](https://mlflow.org/) integration, simply install it in the same virtualenv as OpenKiwi:\n```bash\npip install mlflow\n```\n\n\n## Getting Started\n\nDetailed usage examples and instructions can be found in the [Full Documentation](https://unbabel.github.io/OpenKiwi/index.html).\n\n\n## Pre-trained models\n\nWe provide pre-trained models with the corresponding pre-processed datasets and configuration files.\nYou can easily reproduce our numbers in the WMT 2018 word- and sentence-level tasks by following the [reproduce instructions in the documentation](https://unbabel.github.io/OpenKiwi/reproduce.html).\n\n\n## Contributing\n\nWe welcome contributions to improve OpenKiwi.\nPlease refer to [CONTRIBUTING.md](CONTRIBUTING.md) for quick instructions or to [contributing instructions](https://unbabel.github.io/OpenKiwi/contributing.html) for more detailed instructions on how to set up your development environment.\n\n\n## License\n\nOpenKiwi is Affero GPL licensed. You can see the details of this license in [LICENSE](LICENSE).\n\n\n## Citation\n\nIf you use OpenKiwi, please cite the following paper: [OpenKiwi: An Open Source Framework for Quality Estimation](https://www.aclweb.org/anthology/P19-3020).\n\n```\n@inproceedings{openkiwi,\n    author = {Fábio Kepler and\n              Jonay Trénous and\n              Marcos Treviso and\n              Miguel Vera and\n              André F. T. Martins},\n    title  = {Open{K}iwi: An Open Source Framework for Quality Estimation},\n    year   = {2019},\n    booktitle = {Proceedings of the 57th Annual Meeting of the Association for Computational Linguistics--System Demonstrations},\n    pages  = {117--122},\n    month  = {July},\n    address = {Florence, Italy},\n    url    = {https://www.aclweb.org/anthology/P19-3020},\n    organization = {Association for Computational Linguistics},\n}\n```\n\n\n## References\n\n##### [[1]] [Kreutzer et al. (2015): QUality Estimation from ScraTCH (QUETCH): Deep Learning for Word-level Translation Quality Estimation](http://aclweb.org/anthology/W15-3037)\n[1]:#1-kreutzer-et-al-2015-quality-estimation-from-scratch-quetch-deep-learning-for-word-level-translation-quality-estimation\n\n##### [[2]] [Martins et al. (2016): Unbabel's Participation in the WMT16 Word-Level Translation Quality Estimation Shared Task](http://www.aclweb.org/anthology/W16-2387)\n[2]:#2-martins-et-al-2016-unbabels-participation-in-the-wmt16-word-level-translation-quality-estimation-shared-task\n\n##### [[3]] [Martins et al. (2017): Pushing the Limits of Translation Quality Estimation](http://www.aclweb.org/anthology/Q17-1015)\n[3]:#3-martins-et-al-2017-pushing-the-limits-of-translation-quality-estimation\n\n##### [[4]] [Kim et al. (2017): Predictor-Estimator using Multilevel Task Learning with Stack Propagation for Neural Quality Estimation](http://www.aclweb.org/anthology/W17-4763)\n[4]:#4-kim-et-al-2017-predictor-estimator-using-multilevel-task-learning-with-stack-propagation-for-neural-quality-estimation\n\n##### [[5]] [Wang et al. (2018): Alibaba Submission for WMT18 Quality Estimation Task](http://statmt.org/wmt18/pdf/WMT093.pdf)\n[5]:#5-wang-et-al-2018-alibaba-submission-for-wmt18-quality-estimation-task\n",
    'author': 'AI Research, Unbabel',
    'author_email': 'openkiwi@unbabel.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Unbabel/OpenKiwi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from kiwi import constants as const
from kiwi.data.fieldsets.fieldset import Fieldset


def build_fieldset(base_fieldset):
    source_field = base_fieldset.fields[const.SOURCE]
    target_field = base_fieldset.fields[const.TARGET]

    source_vocab_options = dict(
        min_freq='source_vocab_min_frequency', max_size='source_vocab_size'
    )
    target_vocab_options = dict(
        min_freq='target_vocab_min_frequency', max_size='target_vocab_size'
    )

    extend_vocabs = Fieldset()
    extend_vocabs.add(
        name=const.SOURCE,
        field=source_field,
        file_option_suffix='extend_source_vocab',
        required=None,
        vocab_options=source_vocab_options,
    )
    extend_vocabs.add(
        name=const.TARGET,
        field=target_field,
        file_option_suffix='extend_target_vocab',
        required=None,
        vocab_options=target_vocab_options,
    )

    return extend_vocabs



# -*- coding: utf-8 -*-
"""This defines a generic class for sparse vectors."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import math


class SparseVector(dict):
    """Implementation of a sparse vector using a dictionary."""

    def __init__(self):
        dict.__init__(self)

    def copy(self):
        """Returns a copy of the current vector."""
        vector = SparseVector()
        for key in self:
            vector[key] = self[key]
        return vector

    def as_string(self):
        """Returns a string representation."""
        s = ''
        for key in self:
            s += key + ':' + str(self[key]) + ' '
        return s

    def save(self, f):
        """Save vector to file."""
        for key in self:
            f.write(str(key) + '\t' + str(self[key]) + '\n')

    def load(self, f, dtype=str):
        """Load vector from file."""
        self.clear()
        for line in f:
            fields = line.split('\t')
            key = fields[0]
            value = float(fields[1])
            self[dtype(key)] = value

    def add(self, vector, scalar=1.0):
        """ Adds this vector and a given vector."""
        for key in vector:
            if key in self:
                self[key] += scalar * vector[key]
            else:
                self[key] = scalar * vector[key]

    def scale(self, scalar):
        """Scales this vector by a scale factor."""
        for key in self:
            self[key] *= scalar

    def add_constant(self, scalar):
        """Adds a constant to each element of the vector."""
        for key in self:
            self[key] += scalar

    def squared_norm(self):
        """Computes the squared norm of the vector."""
        return self.dot_product(self)

    def dot_product(self, vector):
        """ Computes the dot product with a given vector.
        Note: this iterates through the self vector, so it may be inefficient
        if the number of nonzeros in self is much larger than the number of
        nonzeros in vector. Hence the function reverts to
        vector.dot_product(self) if that is beneficial."""
        if len(self) > len(vector):
            return vector.dot_product(self)
        value = 0.0
        for key in self:
            if key in vector:
                value += self[key] * vector[key]
        return value

    def normalize(self):
        """ Normalize the vector. Note: if the norm is zero, do nothing."""
        norm = 0.0
        for key in self:
            value = self[key]
            norm += value * value
        norm = math.sqrt(norm)
        if norm > 0.0:
            for key in self:
                self[key] /= norm



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from distutils.util import strtobool

from kiwi.cli.better_argparse import ModelParser
from kiwi.cli.models.quetch import (
    add_data_flags,
    add_predicting_options,
    add_training_data_file_opts,
    add_vocabulary_opts,
)
from kiwi.models.nuqe import NuQE


def add_model_hyper_params_opts(training_parser):
    group = training_parser.add_argument_group('hyper-parameters')
    group.add_argument(
        '--bad-weight',
        type=float,
        default=3.0,
        help='Relative weight for bad labels.',
    )
    group.add_argument(
        '--window-size', type=int, default=3, help='Sliding window size.'
    )
    group.add_argument(
        '--max-aligned',
        type=int,
        default=5,
        help='Max number of alignments between source and target.',
    )
    group.add_argument(
        '--source-embeddings-size',
        type=int,
        default=50,
        help='Word embedding size for source.',
    )
    group.add_argument(
        '--target-embeddings-size',
        type=int,
        default=50,
        help='Word embedding size for target.',
    )
    group.add_argument(
        '--freeze-embeddings',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Freeze embedding weights during training.',
    )
    group.add_argument(
        '--embeddings-dropout',
        type=float,
        default=0.0,
        help='Dropout rate for embedding layers.',
    )

    group.add_argument(
        '--hidden-sizes',
        type=int,
        nargs='+',
        # action='append',
        default=[400, 200, 100, 50],
        help='List of hidden sizes.',
    )
    group.add_argument(
        '--dropout',
        type=float,
        default=0.0,
        help='Dropout rate for linear layers.',
    )
    group.add_argument(
        '--init-type',
        type=str,
        default='uniform',
        choices=[
            'uniform',
            'normal',
            'constant',
            'glorot_uniform',
            'glorot_normal',
        ],
        help='Distribution type for parameters initialization.',
    )
    group.add_argument(
        '--init-support',
        type=float,
        default=0.1,
        help='Parameters are initialized over uniform distribution with '
        'support (-param_init, param_init). Use 0 to not use '
        'initialization.',
    )

    return group


def add_training_options(training_parser):
    add_training_data_file_opts(training_parser)
    add_data_flags(training_parser)
    add_vocabulary_opts(training_parser)
    add_model_hyper_params_opts(training_parser)


def parser_for_pipeline(pipeline):
    if pipeline == 'train':
        return ModelParser(
            'nuqe',
            'train',
            title=NuQE.title,
            options_fn=add_training_options,
            api_module=NuQE,
        )
    if pipeline == 'predict':
        return ModelParser(
            'nuqe',
            'predict',
            title=NuQE.title,
            options_fn=add_predicting_options,
            api_module=NuQE,
        )

    return None



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from collections import defaultdict
from pathlib import Path

import numpy as np
import torch

from kiwi import constants as const
from kiwi import load_model
from kiwi.data import utils
from kiwi.data.builders import build_test_dataset
from kiwi.data.iterators import build_bucket_iterator
from kiwi.data.utils import cross_split_dataset, save_predicted_probabilities
from kiwi.lib import train
from kiwi.lib.utils import merge_namespaces
from kiwi.loggers import tracking_logger

logger = logging.getLogger(__name__)


def run_from_options(options):
    if options is None:
        return

    meta_options = options.meta
    pipeline_options = options.pipeline.pipeline
    model_options = options.pipeline.model
    ModelClass = options.pipeline.model_api

    tracking_run = tracking_logger.configure(
        run_uuid=pipeline_options.run_uuid,
        experiment_name=pipeline_options.experiment_name,
        run_name=pipeline_options.run_name,
        tracking_uri=pipeline_options.mlflow_tracking_uri,
        always_log_artifacts=pipeline_options.mlflow_always_log_artifacts,
    )

    with tracking_run:
        output_dir = train.setup(
            output_dir=pipeline_options.output_dir,
            debug=pipeline_options.debug,
            quiet=pipeline_options.quiet,
        )

        all_options = merge_namespaces(
            meta_options, pipeline_options, model_options
        )
        train.log(
            output_dir,
            config_options=vars(all_options),
            config_file_name='jackknife_config.yml',
        )

        run(
            ModelClass,
            output_dir,
            pipeline_options,
            model_options,
            splits=meta_options.splits,
        )

    teardown(pipeline_options)


def run(ModelClass, output_dir, pipeline_options, model_options, splits):
    model_name = getattr(ModelClass, 'title', ModelClass.__name__)
    logger.info('Jackknifing with the {} model'.format(model_name))

    # Data
    fieldset = ModelClass.fieldset(
        wmt18_format=model_options.__dict__.get('wmt18_format')
    )
    train_set, dev_set = train.retrieve_datasets(
        fieldset, pipeline_options, model_options, output_dir
    )

    test_set = None
    try:
        test_set = build_test_dataset(fieldset, **vars(pipeline_options))
    except ValueError:
        pass
    except FileNotFoundError:
        pass

    device_id = None
    if pipeline_options.gpu_id is not None and pipeline_options.gpu_id >= 0:
        device_id = pipeline_options.gpu_id

    parent_dir = output_dir
    train_predictions = defaultdict(list)
    dev_predictions = defaultdict(list)
    test_predictions = defaultdict(list)
    splitted_datasets = cross_split_dataset(train_set, splits)
    for i, (train_fold, pred_fold) in enumerate(splitted_datasets):

        run_name = 'train_split_{}'.format(i)
        output_dir = Path(parent_dir, run_name)
        output_dir.mkdir(parents=True, exist_ok=True)
        # options.output_dir = str(options.output_dir)

        # Train
        vocabs = utils.fields_to_vocabs(train_fold.fields)

        tracking_run = tracking_logger.start_nested_run(run_name=run_name)
        with tracking_run:
            train.setup(
                output_dir=output_dir,
                seed=pipeline_options.seed,
                gpu_id=pipeline_options.gpu_id,
                debug=pipeline_options.debug,
                quiet=pipeline_options.quiet,
            )

            trainer = train.retrieve_trainer(
                ModelClass,
                pipeline_options,
                model_options,
                vocabs,
                output_dir,
                device_id,
            )

            # Dataset iterators
            train_iter = build_bucket_iterator(
                train_fold,
                batch_size=pipeline_options.train_batch_size,
                is_train=True,
                device=device_id,
            )
            valid_iter = build_bucket_iterator(
                pred_fold,
                batch_size=pipeline_options.valid_batch_size,
                is_train=False,
                device=device_id,
            )

            trainer.run(train_iter, valid_iter, epochs=pipeline_options.epochs)

        # Predict
        predictor = load_model(trainer.checkpointer.best_model_path())
        train_predictions_i = predictor.run(
            pred_fold, batch_size=pipeline_options.valid_batch_size
        )

        dev_predictions_i = predictor.run(
            dev_set, batch_size=pipeline_options.valid_batch_size
        )

        test_predictions_i = None
        if test_set:
            test_predictions_i = predictor.run(
                test_set, batch_size=pipeline_options.valid_batch_size
            )

        torch.cuda.empty_cache()

        for output_name in train_predictions_i:
            train_predictions[output_name] += train_predictions_i[output_name]
            dev_predictions[output_name].append(dev_predictions_i[output_name])
            if test_set:
                test_predictions[output_name].append(
                    test_predictions_i[output_name]
                )

    dev_predictions = average_all(dev_predictions)
    if test_set:
        test_predictions = average_all(test_predictions)

    save_predicted_probabilities(
        parent_dir, train_predictions, prefix=const.TRAIN
    )
    save_predicted_probabilities(parent_dir, dev_predictions, prefix=const.DEV)
    if test_set:
        save_predicted_probabilities(
            parent_dir, test_predictions, prefix=const.TEST
        )

    teardown(pipeline_options)

    return train_predictions


def teardown(options):
    pass


def average_all(predictions):
    for output_name in predictions:
        predictions[output_name] = average_predictions(predictions[output_name])
    return predictions


def average_predictions(ensemble):
    """Average an ensemble of predictions.
    """
    word_level = isinstance(ensemble[0][0], list)
    if word_level:
        sentence_lengths = [len(sentence) for sentence in ensemble[0]]
        ensemble = [
            [word for sentence in predictions for word in sentence]
            for predictions in ensemble
        ]

    ensemble = np.array(ensemble, dtype='float32')
    averaged_predictions = ensemble.mean(axis=0).tolist()

    if word_level:
        averaged_predictions = reshape_by_lengths(
            averaged_predictions, sentence_lengths
        )

    return averaged_predictions


def reshape_by_lengths(sequence, lengths):
    new_sequences = []
    t = 0
    for length in lengths:
        new_sequences.append(sequence[t : t + length])
        t += length
    return new_sequences



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import itertools
from collections import OrderedDict
from pathlib import Path

import configargparse

from kiwi.cli.opts import PathType
from kiwi.lib import train
from kiwi.models.model import Model

parser = configargparse.get_argument_parser('search')

parser.add_argument(
    '-e',
    '--experiment-name',
    required=False,
    help='MLflow will log this run under this experiment name, '
    'which appears as a separate section in the UI. It '
    'will also be used in some messages and files.',
)
parser.add(
    '-c',
    '--config',
    required=True,
    is_config_file=False,
    type=PathType(exists=True),
    help='Load config file from path',
)

group = parser.add_argument_group('models')
group.add_argument('model_name', choices=Model.subclasses.keys())


def get_action(option):
    for action in train.parser._actions:
        if option in train.parser.get_possible_config_keys(action):
            return action
    return None


def split_options(options):
    meta_options = OrderedDict()
    normal_options = []
    for key, value in options.items():
        if isinstance(value, list):
            meta_options[key] = value
        else:
            action = get_action(key)
            normal_options += parser.convert_item_to_command_line_arg(
                action, key, value
            )
    return meta_options, normal_options


def run(options, extra_options):
    config_parser = configargparse.YAMLConfigFileParser()
    config_options = config_parser.parse(Path(options.config).read_text())
    meta, fixed_options = split_options(config_options)

    # Run for each combination of arguments
    fixed_args = [options.model_name] + extra_options
    if options.experiment_name:
        fixed_args += parser.convert_item_to_command_line_arg(
            None, 'experiment-name', options.experiment_name
        )

    meta_keys = meta.keys()
    meta_values = meta.values()
    for values in itertools.product(*meta_values):
        assert len(meta_keys) == len(values)
        run_args = []
        for key, value in zip(meta_keys, values):
            action = get_action(key)
            run_args.extend(
                parser.convert_item_to_command_line_arg(action, key, str(value))
            )
        full_args = fixed_args + run_args + fixed_options
        train.main(full_args)


def main(argv=None, external_options=None):
    raise NotImplementedError('Pipeline not yet supported.')

    # options, extra_options = parser.parse_known_args(args=argv)
    # run(options, extra_options)


if __name__ == '__main__':
    main()



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from functools import partial

from torchtext import data

from kiwi import constants as const
from kiwi.data.corpus import Corpus
from kiwi.data.fields.alignment_field import AlignmentField
from kiwi.data.fields.qe_field import QEField
from kiwi.data.fieldsets.fieldset import Fieldset
from kiwi.data.tokenizers import align_tokenizer, tokenizer


def build_fieldset():
    fs = Fieldset()

    source_vocab_options = dict(
        min_freq='source_vocab_min_frequency', max_size='source_vocab_size'
    )
    target_vocab_options = dict(
        min_freq='target_vocab_min_frequency', max_size='target_vocab_size'
    )

    source_field = QEField(tokenize=tokenizer)
    target_field = QEField(tokenize=tokenizer)
    source_pos = QEField(tokenize=tokenizer)
    target_pos = QEField(tokenize=tokenizer)
    target_tags_field = data.Field(
        tokenize=tokenizer, pad_token=None, unk_token=None
    )

    fs.add(
        name=const.SOURCE,
        field=source_field,
        file_option_suffix='_source',
        required=Fieldset.ALL,
        vocab_options=source_vocab_options,
    )
    fs.add(
        name=const.TARGET,
        field=target_field,
        file_option_suffix='_target',
        required=Fieldset.ALL,
        vocab_options=target_vocab_options,
    )
    fs.add(
        name=const.ALIGNMENTS,
        field=AlignmentField(tokenize=align_tokenizer, use_vocab=False),
        file_option_suffix='_alignments',
        required=Fieldset.ALL,
    )
    fs.add(
        name=const.TARGET_TAGS,
        field=target_tags_field,
        file_option_suffix='_target_tags',
        required=[Fieldset.TRAIN, Fieldset.VALID],
    )

    fs.add(
        name=const.SOURCE_POS,
        field=source_pos,
        file_option_suffix='_source_pos',
        required=None,
    )
    fs.add(
        name=const.TARGET_POS,
        field=target_pos,
        file_option_suffix='_target_pos',
        required=None,
    )

    target_stacked = data.Field(tokenize=tokenizer)
    fs.add(
        name=const.TARGET_STACKED,
        field=target_stacked,
        file_option_suffix='_target_stacked',
        file_reader=partial(Corpus.read_tabular_file, extract_column=1),
        required=None,
    )

    target_parse_heads = data.Field(tokenize=tokenizer, use_vocab=False)
    target_parse_relations = data.Field(tokenize=tokenizer)
    fs.add(
        name=const.TARGET_PARSE_HEADS,
        field=target_parse_heads,
        file_option_suffix='_target_parse',
        file_reader=partial(Corpus.read_tabular_file, extract_column=1),
        required=None,
    )
    fs.add(
        name=const.TARGET_PARSE_RELATIONS,
        field=target_parse_relations,
        file_option_suffix='_target_parse',
        file_reader=partial(Corpus.read_tabular_file, extract_column=2),
        required=None,
    )

    target_ngram_left = data.Field(tokenize=tokenizer)
    target_ngram_right = data.Field(tokenize=tokenizer)
    fs.add(
        name=const.TARGET_NGRAM_LEFT,
        field=target_ngram_left,
        file_option_suffix='_target_ngram',
        file_reader=partial(Corpus.read_tabular_file, extract_column=1),
        required=None,
    )
    fs.add(
        name=const.TARGET_NGRAM_RIGHT,
        field=target_ngram_right,
        file_option_suffix='_target_ngram',
        file_reader=partial(Corpus.read_tabular_file, extract_column=2),
        required=None,
    )

    return fs


#
# def build_test_dataset(options):
#     source_field = QEField(tokenize=tokenizer)
#     target_field = QEField(tokenize=tokenizer)
#     source_pos = QEField(tokenize=tokenizer)
#     target_pos = QEField(tokenize=tokenizer)
#     alignments_field = AlignmentField(
#         tokenize=align_tokenizer, use_vocab=False)
#     target_tags_field = data.Field(
#         tokenize=tokenizer, pad_token=None, unk_token=None
#     )
#     target_parse_heads = data.Field(tokenize=tokenizer, use_vocab=False)
#     target_parse_relations = data.Field(tokenize=tokenizer)
#     target_ngram_left = data.Field(tokenize=tokenizer)
#     target_ngram_right = data.Field(tokenize=tokenizer)
#     target_stacked = data.Field(tokenize=tokenizer)
#
#     fields = {
#         const.SOURCE: source_field,
#         const.TARGET: target_field,
#         const.ALIGNMENTS: alignments_field,
#         const.TARGET_TAGS: target_tags_field
#     }
#
#     test_files = {
#         const.SOURCE: options.test_source,
#         const.TARGET: options.test_target,
#         const.TARGET_TAGS: options.test_target_tags,
#         const.ALIGNMENTS: options.test_alignments,
#     }
#
#     if options.test_target_parse:
#         parse_fields = {
#             const.TARGET_PARSE_HEADS: target_parse_heads,
#             const.TARGET_PARSE_RELATIONS: target_parse_relations,
#         }
#         parse_file_fields = [
#             '',
#             '',
#             '',
#             '',
#             '',
#             const.TARGET_PARSE_HEADS,
#             const.TARGET_PARSE_RELATIONS,
#         ]
#
#     if options.test_target_ngram:
#         ngram_fields = {
#             const.TARGET_NGRAM_LEFT: target_ngram_left,
#             const.TARGET_NGRAM_RIGHT: target_ngram_right,
#         }
#         ngram_file_fields = [
#             '', '', '', '', '', '', '', '', '', '', '', '', '',
#             const.TARGET_NGRAM_LEFT,
#             const.TARGET_NGRAM_RIGHT,
#         ]
#
#     if options.test_target_stacked:
#         stacked_fields = {const.TARGET_STACKED: target_stacked}
#         stacked_file_fields = [const.TARGET_STACKED]
#
#     if options.test_source_pos:
#         fields[const.SOURCE_POS] = source_pos
#         test_files[const.SOURCE_POS] = options.test_source_pos
#
#     if options.test_target_pos:
#         fields[const.TARGET_POS] = target_pos
#         test_files[const.TARGET_POS] = options.test_target_pos
#
#     if options.test_target_parse:
#         test_target_parse_file = options.test_target_parse
#
#     if options.test_target_ngram:
#         test_target_ngram_file = options.test_target_ngram
#
#     if options.test_target_stacked:
#         test_target_stacked_file = options.test_target_stacked
#
#     def filter_len(x):
#         return (
#                    options.source_min_length
#                    <= len(x.source)
#                    <= options.source_max_length
#                ) and (
#                    options.target_min_length
#                    <= len(x.target)
#                    <= options.target_max_length
#                )
#
#     test_examples = Corpus.from_files(fields=fields, files=test_files)
#     if options.test_target_parse:
#         test_examples.paste_fields(
#             Corpus.from_tabular_file(
#                 fields=parse_fields,
#                 file_fields=parse_file_fields,
#                 file_path=test_target_parse_file,
#             )
#         )
#     if options.test_target_ngram:
#         test_examples.paste_fields(
#             Corpus.from_tabular_file(
#                 fields=ngram_fields,
#                 file_fields=ngram_file_fields,
#                 file_path=test_target_ngram_file,
#             )
#         )
#     if options.test_target_stacked:
#         test_examples.paste_fields(
#             Corpus.from_tabular_file(
#                 fields=stacked_fields,
#                 file_fields=stacked_file_fields,
#                 file_path=test_target_stacked_file,
#             )
#         )
#
#     dataset = QEDataset(
#         examples=test_examples,
#         fields=test_examples.dataset_fields,
#         filter_pred=filter_len,
#     )
#
#     return dataset



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import heapq
import logging
import shutil
import threading
from pathlib import Path

import torch

from kiwi import constants as const
from kiwi.data.utils import save_predicted_probabilities

logger = logging.getLogger(__name__)


class EarlyStopException(StopIteration):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Checkpoint:
    """Class for determining whether to evaluate / save the model.
    """

    def __init__(
        self,
        output_dir,
        checkpoint_save=False,
        checkpoint_keep_only_best=0,
        checkpoint_early_stop_patience=0,
        checkpoint_validation_steps=0,
    ):
        """

        Args:
            output_dir (Path): Required if checkpoint_save == True.
            checkpoint_save (bool):
                Save a training snapshot when validation is run.
            checkpoint_keep_only_best:
                Keep only this number of saved snapshots; 0 will keep all.
            checkpoint_early_stop_patience:
                Stop training if evaluation metrics do not improve after /X/
                validations; 0 disables this.
            checkpoint_validation_steps:
                Perform validation every /X/ training batches.
        """
        self.output_directory = Path(output_dir)
        self.validation_steps = checkpoint_validation_steps

        self.early_stop_patience = checkpoint_early_stop_patience

        self.save = checkpoint_save
        self.keep_only_best = checkpoint_keep_only_best
        if self.keep_only_best <= 0:
            self.keep_only_best = float('inf')

        # if self.save and not self.output_directory:
        #     logger.warning('Asked to save training snapshots, '
        #                    'but no output directory was specified.')
        #     self.save = False

        self.main_metric = None

        # This should be kept as a heap
        self.best_stats_summary = []
        self.stats_summary_history = []
        self._last_saved = 0
        self._validation_epoch = 0
        self._best_checkpoint_path = None

    def must_eval(self, epoch=None, step=None):
        if epoch is not None:
            return True
        if step is not None:
            return self.validation_steps and step % self.validation_steps == 0
        return False

    def must_save_best(self, stats):
        if self._validation_epoch <= self.keep_only_best:
            return True
        elif stats > self.worst_stats():
            return True

    def early_stopping(self):
        no_improvement = self._validation_epoch - self._last_saved
        return 0 < self.early_stop_patience <= no_improvement

    def __call__(self, trainer, valid_iterator, epoch=None, step=None):
        if self.must_eval(epoch=epoch, step=step):
            eval_stats_summary = trainer.eval_epoch(valid_iterator)
            eval_stats_summary.log()
            if trainer.scheduler:
                trainer.scheduler.step(eval_stats_summary.main_metric_value())

            saved_path = self.check_in(
                trainer, eval_stats_summary, epoch=epoch, step=step
            )
            if saved_path:
                predictions = trainer.predict(valid_iterator)
                if predictions is not None:
                    save_predicted_probabilities(saved_path, predictions)
            elif self.early_stopping():
                raise EarlyStopException(
                    'Early stopping training after {} validations '
                    'without improvements on the validation set'.format(
                        self.early_stop_patience
                    )
                )

    def check_in(self, trainer, stats, epoch=None, step=None):
        """Saves stat summary and handles checkpoint saving."""
        self._validation_epoch += 1
        self.stats_summary_history.append(stats)
        if self.save:
            if self.must_save_best(stats):
                self._last_saved = self._validation_epoch
                output_path = self.make_output_path(epoch=epoch, step=step)
                path_to_remove = self.push_to_heap(stats, output_path)
                event = trainer.save(output_path)
                self._best_checkpoint_path = output_path
                if path_to_remove:
                    self.remove_snapshot(path_to_remove, event)

                self.save_latest(trainer, saved_best=True)
                return output_path

            output_path = self.save_latest(trainer)
            return output_path
        return None

    def make_output_path(self, epoch=None, step=None):
        if epoch is not None:
            sub_dir = 'epoch_{}'.format(epoch)
        elif step is not None:
            sub_dir = 'step_{}'.format(step)
        else:
            sub_dir = 'epoch_unknown'
        return self.output_directory / sub_dir

    def push_to_heap(self, stats, output_path):
        """Push stats and output path to the heap."""

        path_to_remove = None
        # The second element (`-self._validation_epoch`) serves as a timestamp
        # to ensure that in case of a tie, the earliest model is saved.
        heap_element = (stats, -self._validation_epoch, output_path)
        if self._validation_epoch <= self.keep_only_best:
            heapq.heappush(self.best_stats_summary, heap_element)
        else:
            worst_stat = heapq.heapreplace(
                self.best_stats_summary, heap_element
            )
            path_to_remove = str(worst_stat[2])  # Worst output path
        return path_to_remove

    def remove_snapshot(self, path_to_remove, event=None):
        """Remove snapshot locally and in MLFlow."""

        def _remove_snapshot(path, event, message):
            if event:
                event.wait()
            logger.info(message)
            # ignores directory not empty error
            shutil.rmtree(str(path), ignore_errors=True)
            if event:
                event.clear()

        removal_message = 'Removing previous snapshot: {}'.format(
            path_to_remove
        )

        t = threading.Thread(
            target=_remove_snapshot,
            args=(path_to_remove, event, removal_message),
            daemon=True,
        )
        try:
            t.start()
            return t
        except FileNotFoundError as e:
            logger.exception(e)

    def best_stats_and_path(self):
        if self.best_stats_summary:
            stat, order, path = max(self.best_stats_summary)
            return stat, path
        return None, None

    def best_iteration_path(self):
        _, path = self.best_stats_and_path()
        return path

    def best_stats(self):
        stats, _ = self.best_stats_and_path()
        return stats

    def worst_stats(self):
        if self.best_stats_summary:
            return self.best_stats_summary[0][0]
        else:
            return None

    def best_model_path(self):
        path = self.output_directory / const.BEST_MODEL_FILE
        if path.exists():
            return path
        return self.best_iteration_path()

    def last_model_path(self):
        """Generates the path where the latest model should be saved.
        """
        path = self.output_directory / const.LAST_CHECKPOINT_FOLDER
        temp_path = self.output_directory / const.TEMP_LAST_CHECKPOINT_FOLDER
        return path, temp_path

    def save_latest(self, trainer, saved_best=False):
        """Saves latest checkpoint of the current model.
        In case a model was just saved due to being the best validation, saves a
        pointer instead of the full model. Returns path of saved model.
        """
        output_path, temp_path = self.last_model_path()
        # if best was saved, there's no need to save the entire model twice
        if saved_best:
            temp_path.mkdir(exist_ok=True)
            logging.info('Saving training state to {}'.format(temp_path))
            for file in Path(self._best_checkpoint_path).glob('*'):
                file_path = temp_path / file.name
                torch.save(file, str(file_path))
        else:
            trainer.save(temp_path)

        if output_path.exists():
            t = self.remove_snapshot(output_path)
            t.join()

        logger.info('Moving {} to {}'.format(temp_path, output_path))
        shutil.move(str(temp_path), str(output_path))

        return output_path

    def check_out(self):
        best_path = self.best_iteration_path()
        if best_path:
            self.copy_best_model(best_path, self.output_directory)

    @staticmethod
    def copy_best_model(model_dir, output_dir):
        model_path = model_dir / const.MODEL_FILE
        best_model_path = output_dir / const.BEST_MODEL_FILE
        logging.info('Copying best model to {}'.format(best_model_path))
        shutil.copy(str(model_path), str(best_model_path))
        return best_model_path



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import torch
from torch import nn


class Scorer(nn.Module):
    """Score function for Attention module.
    """

    def __init__(self):
        super().__init__()

    def forward(self, query, keys):
        """Computes Scores for each key given the query.
           args:
                 query:  FloatTensor batch x n
                 keys:   FloatTensor batch x seq_length x m
           ret:
                 scores: FloatTensor batch x seq_length
        """
        raise NotImplementedError


class MLPScorer(Scorer):
    """Implements a score function based on a Multilayer Perceptron.
    """

    def __init__(self, query_size, key_size, layers=2, nonlinearity=nn.Tanh):
        super().__init__()
        layer_list = []
        size = query_size + key_size
        for i in range(layers):
            size_next = size // 2 if i < layers - 1 else 1
            layer_list.append(
                nn.Sequential(nn.Linear(size, size_next), nonlinearity())
            )
            size = size_next
        self.layers = nn.ModuleList(layer_list)

    def forward(self, query, keys):
        layer_in = torch.cat([query.unsqueeze(1).expand_as(keys), keys], dim=-1)
        layer_in = layer_in.reshape(-1, layer_in.size(-1))
        for layer in self.layers:
            layer_in = layer(layer_in)
        out = layer_in.reshape(keys.size()[:-1])
        return out
